// SPDX-License-Identifier: MIT
pragma solidity ^0.8.13;

import "hardhat/console.sol";
import "./Constants.sol";
import "./ConstantsLogic.sol";
import "./IWethzGameState.sol";


contract Logic {
    IWethzGameState public wgs;


    function _isContract(address _addr) internal view returns (bool isContract) {
        uint32 size;
        assembly {
            size := extcodesize(_addr)
        }
        return (size > 0);
    }

    // updates city owner and overlord (if exists) balances
    // uint passAmount - amount passed in case of overlord recursion, otherwise 0
    // uint8 depth - recursion depth
    function _updateCityBalance(uint16 cityId, uint passAmount, uint8 depth) internal {
        uint8 numHomeUnits;
        uint8 numOccupationUnits;

        require(wgs.getCityOwner(cityId) != address(0), "City needs to exist");

        if ((passAmount == 0 && wgs.getCityLastBalanceUpdateTime(cityId) >= block.timestamp) || depth > 3) {
            return;
        }

        numHomeUnits = wgs.getCityUnitIndex(cityId);
        // if more than 3 units, tax collection factor stays the same
        if (numHomeUnits > 3) {
            numHomeUnits = 3;
        }
        numOccupationUnits = wgs.getCityOccupationUnitIndex(cityId);
        // if more than 3 units, tax collection factor stays the same
        if (numOccupationUnits > 3) {
            numOccupationUnits = 3;
        }

        uint revenue = wgs.getTokenPerSec() * (block.timestamp - wgs.getCityLastBalanceUpdateTime(cityId)) + passAmount;
        // apply here additional upgrade effects - START
        // apply here additional upgrade effects - END

        uint16 overlord = wgs.getCityOverlord(cityId);
        if (overlord != 0) {
            // 400 comes from 4 * 100, 4 from unit tax factor and 100 from income factor
            wgs.setBalance(cityId, wgs.getCityBalance(cityId) + (1 + numHomeUnits) * incomeVassal * revenue / 400);
            _updateCityBalance(overlord, (1 + numOccupationUnits) * incomeOverlord * revenue / 400, depth + 1);
        } else {
            // 4 comes from unit tax factor
            wgs.setBalance(cityId, wgs.getCityBalance(cityId) + (1 + numHomeUnits) * revenue / 4);
        }
        wgs.setLastBalanceUpdateTime(cityId, block.timestamp);
    }
}

